from datetime import datetime
from unittest import TestCase

from octue.utils.time import convert_from_posix_time, convert_to_posix_time


class TestTime(TestCase):
    def test_convert_to_posix_time(self):
        """Test that datetime instances can be converted to posix time."""
        self.assertEqual(convert_to_posix_time(datetime(1970, 1, 1)), 0)
        self.assertEqual(convert_to_posix_time(datetime(2000, 1, 1)), 946684800)
        self.assertEqual(convert_to_posix_time(datetime(1940, 1, 1)), -946771200)

    def test_convert_from_posix_time(self):
        """Test that posix timestamps can be converted to datetime instances."""
        self.assertEqual(convert_from_posix_time(0), datetime(1970, 1, 1))
        self.assertEqual(convert_from_posix_time(946684800), datetime(2000, 1, 1))
        self.assertEqual(convert_from_posix_time(-946771200), datetime(1940, 1, 1))



import base64
import json
import logging
import os
from flask import Flask, request

from octue.cloud.pub_sub.service import Service
from octue.exceptions import MissingServiceID
from octue.logging_handlers import apply_log_handler
from octue.resources.service_backends import GCPPubSubBackend
from octue.runner import Runner


logger = logging.getLogger(__name__)
apply_log_handler(logger, log_level=logging.INFO)


DEPLOYMENT_CONFIGURATION_PATH = "deployment_configuration.json"


app = Flask(__name__)


@app.route("/", methods=["POST"])
def index():
    """Receive questions from Google Cloud Run in the form of Google Pub/Sub messages.

    :return (str, int):
    """
    envelope = request.get_json()

    if not envelope:
        return _log_bad_request_and_return_400_response("No Pub/Sub message received.")

    if not isinstance(envelope, dict) or "message" not in envelope:
        return _log_bad_request_and_return_400_response("Invalid Pub/Sub message format.")

    message = envelope["message"]

    if "data" not in message or "attributes" not in message or "question_uuid" not in message["attributes"]:
        return _log_bad_request_and_return_400_response("Invalid Pub/Sub message format.")

    project_name = envelope["subscription"].split("/")[1]
    data = json.loads(base64.b64decode(message["data"]).decode("utf-8").strip())
    question_uuid = message["attributes"]["question_uuid"]
    logger.info("Received question %r.", question_uuid)

    answer_question(project_name, data, question_uuid)
    logger.info("Analysis run and response sent for question %r.", question_uuid)
    return ("", 204)


def _log_bad_request_and_return_400_response(message):
    """Log an error return a bad request (400) response.

    :param str message:
    :return (str, int):
    """
    logger.error(message)
    return (f"Bad Request: {message}", 400)


def answer_question(project_name, data, question_uuid):
    """Answer a question from a service by running the deployed app with the deployment configuration. Either the
    `deployment_configuration_path` should be specified, or the `deployment_configuration`.

    :param str project_name:
    :param dict data:
    :param str question_uuid:
    :return None:
    """
    service_id = os.environ.get("SERVICE_ID")

    if not service_id:
        raise MissingServiceID(
            "The ID for the deployed service is missing - ensure SERVICE_ID is available as an environment variable."
        )

    try:
        with open(DEPLOYMENT_CONFIGURATION_PATH) as f:
            deployment_configuration = json.load(f)

        logger.info("Deployment configuration loaded from %r.", os.path.abspath(DEPLOYMENT_CONFIGURATION_PATH))

    except FileNotFoundError:
        deployment_configuration = {}
        logger.info("Default deployment configuration used.")

    runner = Runner(
        app_src=deployment_configuration.get("app_dir", "."),
        twine=deployment_configuration.get("twine", "twine.json"),
        configuration_values=deployment_configuration.get("configuration_values", None),
        configuration_manifest=deployment_configuration.get("configuration_manifest", None),
        output_manifest_path=deployment_configuration.get("output_manifest", None),
        children=deployment_configuration.get("children", None),
        skip_checks=deployment_configuration.get("skip_checks", False),
        log_level=deployment_configuration.get("log_level", "INFO"),
        handler=deployment_configuration.get("log_handler", None),
        show_twined_logs=deployment_configuration.get("show_twined_logs", False),
        project_name=project_name,
    )

    service = Service(
        service_id=service_id,
        backend=GCPPubSubBackend(project_name=project_name, credentials_environment_variable=None),
        run_function=runner.run,
    )

    service.answer(data=data, question_uuid=question_uuid)



import json
import re
from collections import UserString

from octue.exceptions import InvalidLabelException
from octue.resources.filter_containers import FilterSet


LABEL_PATTERN = re.compile(r"^[a-z0-9][a-z0-9-]*(?<!-)$")


class Label(UserString):
    """A label starts and ends with a character in [A-Za-z0-9] and can contain hyphens e.g. angry-marmaduke

    :param str name: the text of the label
    :return None:
    """

    def __init__(self, name):
        super().__init__(self._clean(name))

    @staticmethod
    def _clean(name):
        """Clean the label name, making sure it is a string and conforms to the label regex pattern.

        :param str name: name of the label to check and clean
        :return str:
        """
        if not isinstance(name, str):
            raise InvalidLabelException("Labels must be expressed as a string.")

        cleaned_name = name.strip()

        if not re.match(LABEL_PATTERN, cleaned_name):
            raise InvalidLabelException(
                f"Invalid label '{cleaned_name}'. Labels must contain only characters 'a-z', 'A-Z', '0-9', and '-'. "
                f"They must not start with '-'."
            )

        return cleaned_name


class LabelSet(set):
    """Class to handle a set of labels.

    :param iter(str|Label)|LabelSet labels: labels to add to the set
    :return None:
    """

    def __init__(self, labels=None):
        # TODO Call the superclass with *args and **kwargs, then update everything to using ResourceBase
        labels = labels or FilterSet()

        # JSON-encoded list of label names, or space-delimited string of label names.
        if isinstance(labels, str):
            try:
                labels = FilterSet(Label(label) for label in json.loads(labels))
            except json.decoder.JSONDecodeError:
                labels = FilterSet(Label(label) for label in labels.strip().split())

        elif isinstance(labels, LabelSet):
            labels = FilterSet(labels)

        # Labels can be some other iterable than a list, but each label must be a Label or string.
        elif hasattr(labels, "__iter__"):
            labels = FilterSet(label if isinstance(label, Label) else Label(label) for label in labels)

        else:
            raise InvalidLabelException(
                "Labels must be expressed as a whitespace-delimited string or an iterable of strings or Label instances."
            )

        super().__init__(labels)

    def add(self, label):
        """Add a label string to the set.

        :param str label: the label to add
        :return None:
        """
        super().add(Label(label))

    def update(self, *labels):
        """Add one or more new label strings to the object labels. New labels will be cleaned and validated.

        :param str *labels: a variable number of string labels
        :return None:
        """
        super().update({Label(label) for label in labels})

    def any_label_starts_with(self, value):
        """Return `True` if any of the labels starts with the value.

        :param str value: value to check
        :return bool:
        """
        return any(label.startswith(value) for label in self)

    def any_label_ends_with(self, value):
        """Return `True` if any of the labels ends with the value.

        :param str value: value to check
        :return bool:
        """
        return any(label.endswith(value) for label in self)

    def any_label_contains(self, value):
        """Return `True` if any of the labels contains the value.

        :param str value: value to check
        :return bool:
        """
        return any(value in label for label in self)



from datetime import datetime, timedelta


def convert_to_posix_time(timestamp):
    """Convert a datetime timestamp to posix time (i.e. seconds since epoch: 1st January 1970).

    :param datetime.datetime timestamp:
    :return float:
    """
    return (timestamp - datetime(1970, 1, 1, tzinfo=timestamp.tzinfo)).total_seconds()


def convert_from_posix_time(posix_timestamp):
    """Convert a posix timestamp to a datetime timestamp.

    :param int|float posix_timestamp:
    :return datetime.datetime:
    """
    return datetime(1970, 1, 1) + timedelta(seconds=posix_timestamp)



from octue import exceptions
from octue.mixins import Filterable
from octue.resources.filter_containers import FilterDict, FilterList, FilterSet
from tests.base import BaseTestCase


class FilterableThing(Filterable):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __eq__(self, other):
        return vars(self) == vars(other)

    def __hash__(self):
        return hash(str(vars(self).items()))


class TestFilterSet(BaseTestCase):
    def test_error_raised_if_any_items_are_not_filterable(self):
        """Test that an error is raised if any items in the FilterSet are not of type Filterable."""
        filter_set = FilterSet([1, 2, 3])

        with self.assertRaises(TypeError):
            filter_set.filter(a__equals=2)

    def test_filter_with_filterables_of_differing_attributes(self):
        """Test filtering with filterables of differing attributes ignores the filterables lacking the filtered-for
        attribute.
        """
        filterables = {FilterableThing(a=3), FilterableThing(b=90), FilterableThing(a=77)}
        filter_set = FilterSet(filterables)

        self.assertEqual(filter_set.filter(a__gt=2), {FilterableThing(a=3), FilterableThing(a=77)})
        self.assertEqual(filter_set.filter(b__equals=90), {FilterableThing(b=90)})
        self.assertEqual(filter_set.filter(b__equals=0), set())

    def test_filter_with_filterables_of_differing_attributes_fails_if_setting_enabled(self):
        """Test filtering with filterables of differing attributes raises an error if any filterables lack the
        filtered-for attribute when `ignore_items_without_attribute` is False.
        """
        filter_set = FilterSet({FilterableThing(a=3), FilterableThing(b=90), FilterableThing(a=77)})

        for kwarg in {"a__gt": 2}, {"b__equals": 90}, {"b__equals": 0}:
            with self.assertRaises(AttributeError):
                filter_set.filter(**kwarg, ignore_items_without_attribute=False)

    def test_filtering_with_multiple_filters(self):
        """Test that multiple filters can be specified in FilterSet.filter at once."""
        filterables = {FilterableThing(a=3, b=2), FilterableThing(a=3, b=99), FilterableThing(a=77)}
        self.assertEqual(FilterSet(filterables).filter(a__equals=3, b__gt=80), {FilterableThing(a=3, b=99)})

    def test_one_fails_if_no_results(self):
        """Test that the `one` method raises an error if there are no results."""
        filterables = FilterSet({FilterableThing(a=3, b=2), FilterableThing(a=3, b=99), FilterableThing(a=77)})

        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException):
            filterables.one(a__equals=10)

    def test_one_fails_if_more_than_one_result(self):
        """Test that the `one` method raises an error if there is more than one result."""
        filterables = FilterSet({FilterableThing(a=3, b=2), FilterableThing(a=3, b=99), FilterableThing(a=77)})

        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException):
            filterables.one(a__equals=3)

    def test_one(self):
        """Test that the `one` method works and returns one result."""
        filterables = FilterSet({FilterableThing(a=3, b=2), FilterableThing(a=3, b=99), FilterableThing(a=77)})
        self.assertEqual(filterables.one(a__equals=77), FilterableThing(a=77))

    def test_ordering_by_a_non_existent_attribute(self):
        """ Ensure an error is raised if ordering is attempted by a non-existent attribute. """
        filter_set = FilterSet([FilterableThing(age=5), FilterableThing(age=4), FilterableThing(age=3)])
        with self.assertRaises(exceptions.InvalidInputException):
            filter_set.order_by("dog-likeness")

    def test_order_by_with_string_attribute(self):
        """ Test ordering a FilterSet by a string attribute returns an appropriately ordered FilterList. """
        cats = [FilterableThing(name="Zorg"), FilterableThing(name="James"), FilterableThing(name="Princess Carolyn")]
        sorted_filter_set = FilterSet(cats).order_by("name")
        self.assertEqual(sorted_filter_set, FilterList([cats[1], cats[2], cats[0]]))

    def test_order_by_with_int_attribute(self):
        """ Test ordering a FilterSet by an integer attribute returns an appropriately ordered FilterList. """
        cats = [FilterableThing(age=5), FilterableThing(age=4), FilterableThing(age=3)]
        sorted_filter_set = FilterSet(cats).order_by("age")
        self.assertEqual(sorted_filter_set, FilterList(reversed(cats)))

    def test_order_by_list_attribute(self):
        """ Test that ordering by list attributes orders by the size of the list. """
        cats = [
            FilterableThing(previous_names=["Scatta", "Catta"]),
            FilterableThing(previous_names=["Kitty"]),
            FilterableThing(previous_names=[]),
        ]
        sorted_filter_set = FilterSet(cats).order_by("previous_names")
        self.assertEqual(sorted_filter_set, FilterList(reversed(cats)))

    def test_order_by_in_reverse(self):
        """ Test ordering in reverse works correctly. """
        cats = [FilterableThing(age=5), FilterableThing(age=3), FilterableThing(age=4)]
        sorted_filter_set = FilterSet(cats).order_by("age", reverse=True)
        self.assertEqual(sorted_filter_set, FilterList([cats[0], cats[2], cats[1]]))


class TestFilterDict(BaseTestCase):
    ANIMALS = FilterDict(
        {
            "cat": FilterableThing(name="Princess Carolyn", age=3, size="small", previous_names=["scatta", "catta"]),
            "dog": FilterableThing(name="Spot", age=90, size="big", previous_names=[]),
            "another_dog": FilterableThing(
                name="Ranger", age=91, size="small", previous_names=["doggo", "oggo", "loggo"]
            ),
        }
    )

    def test_instantiate(self):
        """Test that a FilterDict can be instantiated like a dictionary."""
        filter_dict = FilterDict(a=1, b=3)
        self.assertEqual(filter_dict, {"a": 1, "b": 3})

        filter_dict = FilterDict({"a": 1, "b": 3})
        self.assertEqual(filter_dict, {"a": 1, "b": 3})

        filter_dict = FilterDict(**{"a": 1, "b": 3})
        self.assertEqual(filter_dict, {"a": 1, "b": 3})

    def test_error_raised_when_filtering_if_any_items_are_not_filterable(self):
        """Test that an error is raised if any values in the FilterDict are not of type Filterable."""
        filter_dict = FilterDict({"a": 1, "b": 2, "c": 3})

        with self.assertRaises(TypeError):
            filter_dict.filter(my_attribute__equals=2)

    def test_filter(self):
        """Test that a FilterDict can be filtered on its values when they are all Filterables."""
        filterables = {
            "first-filterable": FilterableThing(my_value=3),
            "second-filterable": FilterableThing(my_value=90),
        }

        filter_dict = FilterDict(filterables)
        self.assertEqual(filter_dict.filter(my_value__equals=90).keys(), {"second-filterable"})
        self.assertEqual(filter_dict.filter(my_value__gt=2), filterables)

    def test_filter_with_filterables_of_differing_attributes(self):
        """Test filtering with filterables of differing attributes ignores the filterables lacking the filtered-for
        attribute.
        """
        filterables = {
            "first-filterable": FilterableThing(a=3),
            "second-filterable": FilterableThing(b=90),
            "third-filterable": FilterableThing(a=77),
        }

        filter_dict = FilterDict(filterables)
        self.assertEqual(filter_dict.filter(a__gt=2).keys(), {"first-filterable", "third-filterable"})
        self.assertEqual(filter_dict.filter(b__equals=90).keys(), {"second-filterable"})
        self.assertEqual(filter_dict.filter(b__equals=0).keys(), set())

    def test_filter_with_filterables_of_differing_attributes_fails_if_setting_enabled(self):
        """Test filtering with filterables of differing attributes raises an error if any filterables lack the
        filtered-for attribute when `ignore_items_without_attribute` is False.
        """
        filterables = {
            "first-filterable": FilterableThing(a=3),
            "second-filterable": FilterableThing(b=90),
            "third-filterable": FilterableThing(a=77),
        }

        filter_dict = FilterDict(filterables)

        for kwarg in {"a__gt": 2}, {"b__equals": 90}, {"b__equals": 0}:
            with self.assertRaises(AttributeError):
                filter_dict.filter(**kwarg, ignore_items_without_attribute=False)

    def test_filter_chaining(self):
        """Test that filters can be chained to filter a FilterDict multiple times."""
        animals_with_age_of_at_least_90 = self.ANIMALS.filter(age__gte=90)
        self.assertEqual({"dog", "another_dog"}, animals_with_age_of_at_least_90.keys())

        animals_with_age_of_at_least_90_and_size_small = animals_with_age_of_at_least_90.filter(size__equals="small")
        self.assertEqual(animals_with_age_of_at_least_90_and_size_small.keys(), {"another_dog"})

    def test_filtering_with_multiple_filters(self):
        """Test that multiple filters can be specified in FilterDict.filter at once."""
        self.assertEqual(self.ANIMALS.filter(size__equals="small", age__lt=5), {"cat": self.ANIMALS["cat"]})

    def test_one_fails_if_no_results(self):
        """Test that the `one` method raises an error if there are no results."""
        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException):
            self.ANIMALS.one(age__equals=10)

    def test_one_fails_if_more_than_one_result(self):
        """Test that the `one` method raises an error if there is more than one result."""
        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException):
            self.ANIMALS.one(size__equals="small")

    def test_one(self):
        """Test that the `one` method works and returns one result."""
        self.assertEqual(self.ANIMALS.one(age__equals=91), ("another_dog", self.ANIMALS["another_dog"]))

    def test_ordering_by_a_non_existent_attribute(self):
        """Ensure an error is raised if ordering is attempted by a non-existent attribute."""
        with self.assertRaises(exceptions.InvalidInputException):
            self.ANIMALS.order_by("dog-likeness")

    def test_order_by_with_string_attribute(self):
        """Test that ordering a FilterDict by a string attribute returns an appropriately ordered FilterList."""
        self.assertEqual(
            self.ANIMALS.order_by("name"),
            FilterList(
                (
                    ("cat", self.ANIMALS["cat"]),
                    ("another_dog", self.ANIMALS["another_dog"]),
                    ("dog", self.ANIMALS["dog"]),
                )
            ),
        )

    def test_order_by_with_int_attribute(self):
        """ Test ordering a FilterDict by an integer attribute returns an appropriately ordered FilterList. """
        self.assertEqual(
            self.ANIMALS.order_by("age"),
            FilterList(
                (
                    ("cat", self.ANIMALS["cat"]),
                    ("dog", self.ANIMALS["dog"]),
                    ("another_dog", self.ANIMALS["another_dog"]),
                )
            ),
        )

    def test_order_by_list_attribute(self):
        """Test that ordering by list attributes orders members alphabetically by the first element of each list."""
        self.assertEqual(
            self.ANIMALS.order_by("previous_names"),
            FilterList(
                (
                    ("dog", self.ANIMALS["dog"]),
                    ("another_dog", self.ANIMALS["another_dog"]),
                    ("cat", self.ANIMALS["cat"]),
                )
            ),
        )

    def test_order_by_in_reverse(self):
        """ Test ordering in reverse works correctly. """
        self.assertEqual(
            self.ANIMALS.order_by("age", reverse=True),
            FilterList(
                (
                    ("another_dog", self.ANIMALS["another_dog"]),
                    ("dog", self.ANIMALS["dog"]),
                    ("cat", self.ANIMALS["cat"]),
                )
            ),
        )



import json
import logging
import time
import uuid
from concurrent.futures import TimeoutError
import google.api_core
import google.api_core.exceptions
from google.api_core import retry
from google.cloud import pubsub_v1

from octue import exceptions
from octue.cloud.credentials import GCPCredentialsManager
from octue.cloud.pub_sub import Subscription, Topic
from octue.exceptions import FileLocationError
from octue.mixins import CoolNameable
from octue.resources.manifest import Manifest
from octue.utils.encoders import OctueJSONEncoder


logger = logging.getLogger(__name__)


OCTUE_NAMESPACE = "octue.services"
ANSWERS_NAMESPACE = "answers"


# Switch message batching off by setting max_messages to 1. This minimises latency and is recommended for
# microservices publishing single messages in a request-response sequence.
BATCH_SETTINGS = pubsub_v1.types.BatchSettings(max_bytes=10 * 1000 * 1000, max_latency=0.01, max_messages=1)


def create_custom_retry(timeout):
    """Create a custom `Retry` object specifying that the given Google Cloud request should retry for the given amount
    of time for the given exceptions.

    :param float timeout:
    :return google.api_core.retry.Retry:
    """
    return retry.Retry(
        maximum=timeout / 4,
        deadline=timeout,
        predicate=google.api_core.retry.if_exception_type(
            google.api_core.exceptions.NotFound,
            google.api_core.exceptions.Aborted,
            google.api_core.exceptions.DeadlineExceeded,
            google.api_core.exceptions.InternalServerError,
            google.api_core.exceptions.ResourceExhausted,
            google.api_core.exceptions.ServiceUnavailable,
            google.api_core.exceptions.Unknown,
            google.api_core.exceptions.Cancelled,
        ),
    )


class Service(CoolNameable):
    """A Twined service that can be used in two modes:
    * As a server accepting questions (input values and manifests), running them through its app, and responding to the
    requesting service with the results of the analysis.
    * As a requester of answers from another Service in the above mode.

    Services communicate entirely via Google Pub/Sub and can ask and/or respond to questions from any other Service that
    has a corresponding topic on Google Pub/Sub.

    :param octue.resources.service_backends.ServiceBackend backend:
    :param str|None service_id:
    :param callable|None run_function:
    :return None:
    """

    def __init__(self, backend, service_id=None, run_function=None):
        if service_id is None:
            self.id = str(uuid.uuid4())
        elif not service_id:
            raise ValueError(f"service_id should be None or a non-falsey value; received {service_id!r} instead.")
        else:
            self.id = service_id

        self.backend = backend
        self.run_function = run_function

        credentials = GCPCredentialsManager(backend.credentials_environment_variable).get_credentials()
        self.publisher = pubsub_v1.PublisherClient(credentials=credentials, batch_settings=BATCH_SETTINGS)
        self.subscriber = pubsub_v1.SubscriberClient(credentials=credentials)
        super().__init__()

    def __repr__(self):
        return f"<{type(self).__name__}({self.name!r})>"

    def serve(self, timeout=None, delete_topic_and_subscription_on_exit=False):
        """Start the Service as a server, waiting to accept questions from any other Service using Google Pub/Sub on
        the same Google Cloud Platform project. Questions are responded to asynchronously."""
        topic = Topic(name=self.id, namespace=OCTUE_NAMESPACE, service=self)
        topic.create(allow_existing=True)

        subscription = Subscription(
            name=self.id, topic=topic, namespace=OCTUE_NAMESPACE, service=self, expiration_time=None
        )
        subscription.create(allow_existing=True)

        future = self.subscriber.subscribe(subscription=subscription.path, callback=self.receive_question_then_answer)
        logger.debug("%r is waiting for questions.", self)

        with self.subscriber:
            try:
                future.result(timeout=timeout)
            except (TimeoutError, KeyboardInterrupt):
                future.cancel()

            if delete_topic_and_subscription_on_exit:
                topic.delete()
                subscription.delete()

    def receive_question_then_answer(self, question):
        """Receive a question, acknowledge it, then answer it."""
        logger.info("%r received a question.", self)
        data = json.loads(question.data.decode())
        question_uuid = question.attributes["question_uuid"]
        question.ack()
        self.answer(data, question_uuid)

    def answer(self, data, question_uuid, timeout=30):
        """Answer a question (i.e. run the Service's app to analyse the given data, and return the output values to the
        asker). Answers are published to a topic whose name is generated from the UUID sent with the question, and are
        in the format specified in the Service's Twine file.
        """
        topic = Topic(
            name=".".join((self.id, ANSWERS_NAMESPACE, question_uuid)), namespace=OCTUE_NAMESPACE, service=self
        )
        analysis = self.run_function(input_values=data["input_values"], input_manifest=data["input_manifest"])

        if analysis.output_manifest is None:
            serialised_output_manifest = None
        else:
            serialised_output_manifest = analysis.output_manifest.serialise(to_string=True)

        self.publisher.publish(
            topic=topic.path,
            data=json.dumps(
                {"output_values": analysis.output_values, "output_manifest": serialised_output_manifest},
                cls=OctueJSONEncoder,
            ).encode(),
            retry=create_custom_retry(timeout),
        )
        logger.info("%r responded on topic %r.", self, topic.path)

    def ask(self, service_id, input_values, input_manifest=None):
        """Ask a serving Service a question (i.e. send it input values for it to run its app on). The input values must
        be in the format specified by the serving Service's Twine file. A single-use topic and subscription are created
        before sending the question to the serving Service - the topic is the expected publishing place for the answer
        from the serving Service when it comes, and the subscription is set up to subscribe to this.
        """
        if (input_manifest is not None) and (not input_manifest.all_datasets_are_in_cloud):
            raise FileLocationError(
                "All datasets of the input manifest and all files of the datasets must be uploaded to the cloud before "
                "asking a service to perform an analysis upon them. The manifest must then be updated with the new "
                "cloud locations."
            )

        question_topic = Topic(name=service_id, namespace=OCTUE_NAMESPACE, service=self)
        if not question_topic.exists():
            raise exceptions.ServiceNotFound(f"Service with ID {service_id!r} cannot be found.")

        question_uuid = str(int(uuid.uuid4()))

        response_topic_and_subscription_name = ".".join((service_id, ANSWERS_NAMESPACE, question_uuid))
        response_topic = Topic(name=response_topic_and_subscription_name, namespace=OCTUE_NAMESPACE, service=self)
        response_topic.create(allow_existing=False)

        response_subscription = Subscription(
            name=response_topic_and_subscription_name,
            topic=response_topic,
            namespace=OCTUE_NAMESPACE,
            service=self,
        )
        response_subscription.create(allow_existing=False)

        if input_manifest is not None:
            input_manifest = input_manifest.serialise(to_string=True)

        future = self.publisher.publish(
            topic=question_topic.path,
            data=json.dumps({"input_values": input_values, "input_manifest": input_manifest}).encode(),
            question_uuid=question_uuid,
        )
        future.result()

        logger.debug("%r asked question to %r service. Question UUID is %r.", self, service_id, question_uuid)
        return response_subscription, question_uuid

    def wait_for_answer(self, subscription, timeout=30):
        """Wait for an answer to a question on the given subscription, deleting the subscription and its topic once
        the answer is received.

        :param octue.cloud.pub_sub.subscription.Subscription subscription: the subscription for the question's answer
        :param float timeout: how long to wait for an answer before raising a TimeoutError
        :raise TimeoutError: if the timeout is exceeded
        :return dict: dictionary containing the keys "output_values" and "output_manifest"
        """
        start_time = time.perf_counter()
        no_message = True
        attempt = 1

        with self.subscriber:

            try:
                while no_message:
                    logger.debug("Pulling messages from Google Pub/Sub: attempt %d.", attempt)

                    pull_response = self.subscriber.pull(
                        request={"subscription": subscription.path, "max_messages": 1},
                        retry=create_custom_retry(timeout),
                    )

                    try:
                        answer = pull_response.received_messages[0]
                        no_message = False

                    except IndexError:
                        logger.debug("Google Pub/Sub pull response timed out early.")
                        attempt += 1

                        if (time.perf_counter() - start_time) > timeout:
                            raise TimeoutError(
                                f"No answer received from topic {subscription.topic.path!r} after {timeout} seconds.",
                            )

            finally:
                try:
                    self.subscriber.acknowledge(request={"subscription": subscription.path, "ack_ids": [answer.ack_id]})
                    logger.debug("%r received a response to question on topic %r.", self, subscription.topic.path)
                except UnboundLocalError:
                    pass

                subscription.delete()
                subscription.topic.delete()

        data = json.loads(answer.message.data.decode())

        if data["output_manifest"] is None:
            output_manifest = None
        else:
            output_manifest = Manifest.deserialise(data["output_manifest"], from_string=True)

        return {"output_values": data["output_values"], "output_manifest": output_manifest}
